############################################################################
#                                                                          #
# Imports & Constants                                                      #
#                                                                          #
############################################################################

import gpudb
import csv
import sys

HOST = "127.0.0.1"
PORT = "9191"
CSV = "nyc_neighborhood.csv"

OPTION_NO_ERROR = {"no_error_if_not_exists": "true"}

TABLE_NYC_N = "nyc_neighborhood"
TABLE_TAXI = "nyctaxi"
TABLE_TAXI_EW = TABLE_TAXI + "_edges_weights_wkt"

JOIN_TAXI = "taxi_tables_joined"

GRAPH_T = TABLE_TAXI + "_graph_wkt"
GRAPH_T_MRSOLVED = GRAPH_T + "_multiple_routing_solved"

############################################################################
#                                                                          #
# Table Setup Function                                                     #
#                                                                          #
############################################################################

def table_setup():
    """ Setup necessary source tables for graph solver examples. """

    print("")
    print("===========")
    print("TABLE SETUP")
    print("===========")
    print("")

    # Clear any related tables in case they already exist
    kinetica.clear_table(table_name=JOIN_TAXI, options=OPTION_NO_ERROR)
    kinetica.clear_table(table_name=TABLE_NYC_N, options=OPTION_NO_ERROR)
    kinetica.clear_table(table_name=TABLE_TAXI_EW, options=OPTION_NO_ERROR)

    # Define the NYC neighborhood table type
    nycn_columns = [
        ["gid", "int"],
        ["geom", "string", "wkt"],
        ["CTLabel", "string", "char16"],
        ["BoroCode", "string", "char16"],
        ["BoroName", "string", "char16"],
        ["CT2010", "string", "char16"],
        ["BoroCT2010", "string", "char16"],
        ["CDEligibil", "string", "char16"],
        ["NTACode", "string", "char16"],
        ["NTAName", "string", "char64"],
        ["PUMA", "string", "char16"],
        ["Shape_Leng", "double"],
        ["Shape_Area", "double"]
    ]

    # Create the NYC neighborhood table
    try:
        table_nycn_obj = gpudb.GPUdbTable(
            _type=nycn_columns,
            name=TABLE_NYC_N,
            db=kinetica,
            options={"is_replicated": "true"}
        )
        print("{} table object successfully created.".format(TABLE_NYC_N))
    except gpudb.GPUdbException as e:
        print(
            "{} table object creation failure: {}".format(TABLE_NYC_N, str(e))
        )

    # Increase the field size limit because CSV contains large field values
    csv.field_size_limit(sys.maxsize)

    # Insert records from a local CSV file into the NYC neighborhood table
    print("Creating records from the " + CSV + " file.")
    nycn_data = csv.reader(open(CSV))
    nycn_data.next()
    nycn_records = []
    for record in nycn_data:
        record_data = [int(record[0]), record[1], record[2], record[3],
                       record[4], record[5], record[6], record[7], record[8],
                       record[9], record[10], float(record[11]),
                       float(record[12])]
        nycn_records.append(record_data)
    print(
        "Inserting records into {} table. This may take a "
        "while.".format(TABLE_NYC_N)
    )
    table_nycn_obj.insert_records(nycn_records)
    print(
        "Number of records inserted into the {} "
        "table: {}".format(TABLE_NYC_N, table_nycn_obj.size())
    )
    print("")

    # Check to see if 'nyctaxi' table exists
    print("Checking to see if table {} exists.".format(TABLE_TAXI))
    nyctaxi_status = kinetica.has_table(
        table_name=TABLE_TAXI
    )['table_exists']
    if nyctaxi_status:
        print("Table {} exists.".format(TABLE_TAXI))
    else:
        print(
            "Table {} does not exist. Please ingest the table (in "
            "GAdmin, Cluster > Demo > NYC Taxi > Load Sample Data) and try "
            "again.".format(TABLE_TAXI))
        sys.exit(1)
    print("")

    # Join the TABLE_TAXI table to the TABLE_NYC_N table using STXY_CONTAINS
    # to filter out data that could skew the graph
    print(
        "Joining {} to {} to filter out data that could skew the taxi "
        "graphs.".format(TABLE_TAXI, JOIN_TAXI)
    )
    join_taxi_tables_response = kinetica.create_join_table(
        join_table_name=JOIN_TAXI,
        table_names=[TABLE_TAXI + " as t", TABLE_NYC_N + " as n"],
        column_names=[
            "CONCAT(CHAR32(pickup_longitude), CHAR32(pickup_latitude)) as pickup_name",
            "t.pickup_longitude", 
            "t.pickup_latitude",
            "HASH(t.pickup_longitude + t.pickup_latitude) as pickup_id",
            "CONCAT(CHAR32(dropoff_longitude), CHAR32(dropoff_latitude)) as dropoff_name",
            "t.dropoff_longitude", 
            "t.dropoff_latitude",
            "HASH(t.dropoff_longitude + t.dropoff_latitude) as dropoff_id",
            "t.total_amount"
        ],
        expressions=[
            "(STXY_CONTAINS(n.geom, t.pickup_longitude, t.pickup_latitude)) AND"
            "(STXY_CONTAINS(n.geom, t.dropoff_longitude, t.dropoff_latitude)) "
        ]
    )["status_info"]["status"]
    print("{} view created: {}".format(JOIN_TAXI, join_taxi_tables_response))
    print("")

    # Create a projection to contain the graph edges (based on EDGE_WKTLINE)
    print(
        "Creating a projection from {} to contain the {} edges.".format(
            JOIN_TAXI, GRAPH_T
        )
    )
    edges_wkt_response = kinetica.create_projection(
        table_name=JOIN_TAXI,
        projection_name=TABLE_TAXI_EW,
        column_names=[
            "REMOVE_NULLABLE(ST_MAKELINE(ST_MAKEPOINT(pickup_longitude, pickup_latitude), ST_MAKEPOINT(dropoff_longitude, dropoff_latitude))) AS tripwkt",
            "total_amount"
        ],
        options={}
    )["status_info"]["status"]
    print(
        "{} projection created: {}".format(
            TABLE_TAXI_EW, edges_wkt_response
        )
    )

# end table_setup()

############################################################################
#                                                                          #
# Graph Setup Function                                                     #
#                                                                          #
############################################################################

def graph_setup():
    """ Setup graphs to be used in solver examples. """

    print("")
    print("===========")
    print("GRAPH SETUP")
    print("===========")
    print("")

    # Create a graph from TABLE_TAXI_EW using WKTLINE
    print("Creating {}".format(GRAPH_T))
    create_t_graph_response = kinetica.create_graph(
        graph_name=GRAPH_T,
        directed_graph=False,
        nodes=[],
        edges=[
            TABLE_TAXI_EW + ".tripwkt AS EDGE_WKTLINE"
        ],
        weights=[
            TABLE_TAXI_EW + ".tripwkt AS WEIGHTS_EDGE_WKTLINE",
            TABLE_TAXI_EW + ".total_amount AS WEIGHTS_VALUESPECIFIED"
        ],
        restrictions=[],
        options={
            "recreate": "true",
            "merge_tolerance": "0.01"
        }
    )
    print(
        "{} created: {}".format(
          GRAPH_T, create_t_graph_response["status_info"]["status"])
    )
    print("Number of nodes: {}".format(create_t_graph_response["num_nodes"]))
    print("Number of edges: {}".format(create_t_graph_response["num_edges"]))

# end graph_setup()

############################################################################
#                                                                          #
# Multiple Routing Solving Function                                        #
#                                                                          #
############################################################################

def multiple_routing_example():
    """Demonstrate creating and solving a graph using the MULTIPLE_ROUTING
    method (Travelling Salesman).
    """

    print("")
    print("===============================")
    print("MULTIPLE ROUTING SOLVER EXAMPLE")
    print("===============================")
    print("")

    # Clear any related tables in case they already exist
    kinetica.clear_table(
        table_name=GRAPH_T_MRSOLVED,
        options=OPTION_NO_ERROR
    )

    # Solve GRAPH_T for Multiple Routing using the given source node and
    # routing to the given destination nodes
    print("Solving {} using MULTIPLE_ROUTING solver type.".format(GRAPH_T))
    source_node = "POINT(-73.98438262939453 40.76493835449219)"  # node 2
    destination_nodes = [
        "POINT(-73.97122955322266 40.74700927734375)",  # node 39
        "POINT(-73.97740173339844 40.77263641357422)"  # node 85
    ]
    solve_t1_mrgraph_response = kinetica.solve_graph(
        graph_name=GRAPH_T,
        solver_type="MULTIPLE_ROUTING",
        source_nodes=[source_node], 
        destination_nodes=destination_nodes,
        solution_table=GRAPH_T_MRSOLVED,
        options={"export_solve_results": "true"}
    )["result_per_destination_node"][0]
    print(
        "Cost for source node {} to visit destination nodes "
        "{}: ${}".format(
            source_node,
            destination_nodes,
            (round(solve_t1_mrgraph_response, 2))
        )
    )
    print("")

# end multiple_routing_example()


if __name__ == '__main__':
    # Establish connection with a locally-running instance of Kinetica
    kinetica = gpudb.GPUdb(encoding='BINARY', host=HOST, port=PORT)

    # Execute defined functions
    table_setup()
    graph_setup()
    multiple_routing_example()
